"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.saveManigest = exports.getManifest = exports.detectChanges = exports.snapshot = void 0;
const path = require("path");
const chalk = require("chalk");
const pLimit = require("p-limit");
const crypto = require("crypto");
const diff_utils_1 = require("@paperclip-ui/diff-utils");
const fsa = require("fs-extra");
const utils_1 = require("./utils");
const ora = require("ora");
const mime = require("mime-types");
const state_1 = require("./state");
const PNG = require("pngjs").PNG;
const pixelmatch = require("pixelmatch");
// TODO - need to pull from config
const WINDOW_SIZES = [{ width: 1400, height: 768 }];
/**
 * Generates new snapshot from current branch
 */
const snapshot = (provider) => (force) => __awaiter(void 0, void 0, void 0, function* () {
    const status = yield provider.git.status();
    if (status.files.length && !force) {
        return (0, utils_1.logWarn)(`You need to commit your file changes before creating a snapshot. You may add ${chalk.bold("--force")} to ignore this warning.`);
    }
    const latestCommit = (yield provider.git.log()).latest.hash;
    const currentBranch = yield (yield provider.git.branch()).current;
    const spinner = ora(`Generating base ${chalk.bold(currentBranch)} snapshots...`).start();
    yield saveScreenshots(provider)(latestCommit, currentBranch);
    spinner.stop();
});
exports.snapshot = snapshot;
const detectChanges = (provider) => (branch) => __awaiter(void 0, void 0, void 0, function* () {
    const { git, cwd, gitDir } = provider;
    const currentBranch = yield (yield git.branch()).current;
    const deltaBranch = branch || currentBranch;
    let manifest = (0, exports.getManifest)(gitDir);
    const deltaCommit = manifest.commitAliases[deltaBranch];
    let spinner = ora(`Generating latest ${chalk.bold(currentBranch)} snapshots...`).start();
    const version = currentBranch;
    manifest = yield saveScreenshots(provider)(version, version + "@latest");
    yield spinner.stop();
    yield timeout(500);
    (0, utils_1.logInfo)(`Diffing against commit sha ${chalk.bold(deltaCommit)}...`);
    const snapshot = manifest.projectSnapshots.find(snapshot => snapshot.version === version);
    const deltaSnapshot = manifest.projectSnapshots.find(snapshot => snapshot.version === deltaCommit);
    // return {
    //   manifest,
    //   currentVersion: snapshot.version,
    //   deltaVersion: deltaSnapshot.version
    // };
    const snapshotPairs = getSnapshotPairs(snapshot, deltaSnapshot, manifest);
    for (const hash in snapshotPairs) {
        const screenshotPairs = snapshotPairs[hash];
        for (const [aid, bid] of screenshotPairs) {
            // it's a new frame
            if (!bid) {
                continue;
            }
            const apath = (0, state_1.getScreenshotPath)(gitDir, aid);
            const bpath = (0, state_1.getScreenshotPath)(gitDir, bid);
            const info = yield diffImages(apath, bpath);
            const diff = {
                id: md5(aid + bid),
                changeCount: info.changeCount,
                screnshotId: aid,
                otherScreenshotId: bid
            };
            yield saveDiffScreenshot(gitDir, diff.id, info);
            manifest = (0, state_1.addDiffToManifest)(aid, diff, manifest);
        }
    }
    (0, exports.saveManigest)(gitDir, manifest);
    return {
        manifest,
        currentVersion: snapshot.version,
        deltaVersion: deltaSnapshot.version
    };
});
exports.detectChanges = detectChanges;
const getSnapshotPairs = (a, b, manifest) => {
    const framePairs = {};
    const aframes = manifest.frameSnapshots.filter(v => a.frameIds.includes(v.id));
    const bframes = manifest.frameSnapshots.filter(v => b.frameIds.includes(v.id));
    for (const aframe of aframes) {
        const bframe = bframes.find(frame => frame.hash === aframe.hash);
        const screenshotPairs = [];
        const ascreenshots = manifest.screenshots.filter(screenshot => aframe.screenshotIds.includes(screenshot.id));
        for (const ascreenshot of ascreenshots) {
            const pairIds = [ascreenshot.id];
            const bscreenshot = bframe &&
                manifest.screenshots.find(screenshot => bframe.screenshotIds.includes(screenshot.id) &&
                    (0, state_1.screenshotEquals)(ascreenshot, screenshot));
            if (bscreenshot) {
                pairIds.push(bscreenshot.id);
            }
            screenshotPairs.push(pairIds);
        }
        framePairs[aframe.hash] = screenshotPairs;
    }
    return framePairs;
};
/**
 */
/**
 */
const saveDiffScreenshot = (gitDir, id, { diff }) => {
    const filePath = (0, state_1.getScreenshotPath)(gitDir, id);
    fsa.writeFileSync(filePath, PNG.sync.write(diff));
};
/**
 */
const timeout = (ms) => new Promise(resolve => setTimeout(resolve, ms));
/**
 */
const getManifestPath = (cwd) => path.join(cwd, state_1.DATA_DIR, state_1.MANIFEST_FILE_NAME);
const getManifest = (gitDir) => {
    const filePath = getManifestPath(gitDir);
    if (fsa.existsSync(filePath)) {
        const manifest = JSON.parse(fsa.readFileSync(filePath, "utf-8"));
        // just dump it
        if (manifest.version === state_1.MANIFEST_VERSION) {
            return manifest;
        }
    }
    return {
        version: state_1.MANIFEST_VERSION,
        commitAliases: {},
        frameSnapshots: [],
        screenshots: [],
        diffs: [],
        projectSnapshots: []
    };
};
exports.getManifest = getManifest;
/**
 */
const saveScreenshots = ({ gitDir, cwd, browser, git }) => (version, tag) => __awaiter(void 0, void 0, void 0, function* () {
    yield fsa.mkdirp(getScreenshotsDir(gitDir));
    const limit = pLimit(10);
    const allFrameSnapshots = [];
    const allScreenshots = [];
    const resolveAsset = (filePath) => {
        // console.log(`data:${mime.lookup(filePath)};base64,${fsa.readFileSync(filePath).toString("base64")}`);
        return `data:${mime.lookup(filePath)};base64,${fsa
            .readFileSync(filePath)
            .toString("base64")}`;
    };
    yield (0, diff_utils_1.eachFrame)(".", { cwd, resolveAsset }, ({ id: frameHash, html, title, filePath }) => __awaiter(void 0, void 0, void 0, function* () {
        return limit(() => __awaiter(void 0, void 0, void 0, function* () {
            const browserVersion = yield browser.version();
            const screenshotIds = [];
            for (const size of WINDOW_SIZES) {
                const hash = md5(frameHash + browserVersion + size.width + size.height);
                const id = md5(version + hash);
                const filePath = (0, state_1.getScreenshotPath)(gitDir, id);
                const page = yield browser.newPage();
                page.setViewport(size);
                yield page.setContent(html);
                // stop text input focus
                yield page.focus("body");
                // wait for content to load
                yield page.waitForNetworkIdle();
                // TODO - wait for transition end
                yield page.screenshot({
                    path: filePath
                });
                screenshotIds.push(id);
                allScreenshots.push({
                    id,
                    size,
                    browser: browserVersion,
                    diffIds: []
                });
            }
            allFrameSnapshots.push({
                id: md5(version + frameHash),
                hash: frameHash,
                title,
                sourceFilePath: path.relative(gitDir, filePath),
                screenshotIds
            });
        }));
    }));
    const manifest = yield (0, state_1.addFrameSnapshots)({ version, frameIds: allFrameSnapshots.map(snapshot => snapshot.id) }, allFrameSnapshots, allScreenshots, gitDir, version, tag, (0, exports.getManifest)(gitDir));
    (0, exports.saveManigest)(gitDir, manifest);
    return manifest;
});
/**
 */
const diffImages = (snapshotFilePath, deltaSnapshotPath) => {
    const img1 = PNG.sync.read(fsa.readFileSync(snapshotFilePath));
    const img2 = PNG.sync.read(fsa.readFileSync(deltaSnapshotPath));
    const { width, height } = img1;
    const diff = new PNG({ width, height });
    const changeCount = pixelmatch(img1.data, img2.data, diff.data, width, height, {
        threshold: 0.2
    });
    return { diff, changeCount };
};
/**
 */
const saveManigest = (gitDir, manifest) => {
    const filePath = getManifestPath(gitDir);
    fsa.writeFileSync(filePath, JSON.stringify(manifest, null, 2));
};
exports.saveManigest = saveManigest;
/**
 */
const getScreenshotsDir = (gitDir) => path.join(gitDir, state_1.SCREENSHOTS_DIR);
const md5 = (value) => {
    return crypto
        .createHash("md5")
        .update(value)
        .digest("hex");
};
